from cucu import helpers
from cucu.utils import take_saw_element_screenshot


def find_image(ctx, name, index=0):
    """
    find an input on screen by fuzzy matching on the name provided and the
    target element:

        * <img>

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired image on screen
      index(str):  the index of the image if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()

    name = name.replace('"', '\\"')
    images = ctx.browser.css_find_elements(f'img[alt="{name}"')

    if index >= len(images):
        element = None
    else:
        element = images[index]

    take_saw_element_screenshot(ctx, "image", name, index, element)

    return element


helpers.define_should_see_thing_with_name_steps(
    "image with the alt text", find_image
)



import shutil
import sys
import traceback
from datetime import datetime, timezone
from pathlib import Path
from xml.sax.saxutils import escape as escape_

import jinja2
from playhouse import shortcuts

import cucu.db as db
from cucu import format_gherkin_table, logger
from cucu.ansi_parser import parse_log_to_html
from cucu.config import CONFIG
from cucu.utils import behave_filepath_to_cucu_logpath, ellipsize_filename


def escape(data):
    if data is None:
        return None

    return escape_(data, {'"': "&quot;"}).rstrip()


def process_tags(element):
    """
    process tags in the element provided (scenario or feature) and basically
    convert the tags to a simple @xxx representation.
    """
    prepared_tags = []

    if "tags" not in element:
        return

    for tag in element["tags"]:
        tag = f"@{tag}"

        # process custom tag handlers
        tag_handlers = CONFIG["__CUCU_HTML_REPORT_TAG_HANDLERS"].items()
        for regex, handler in tag_handlers:
            if regex.match(tag):
                tag = handler(tag)

        prepared_tags.append(tag)

    element["tags"] = " ".join(prepared_tags)


def urlencode(string):
    """
    handles encoding specific characters in the names of features/scenarios
    so they can be used in a URL. NOTICE: we're not handling spaces since
    the browser handles those already.

    """
    return string.replace('"', "%22").replace("'", "%27").replace("#", "%23")


def left_pad_zeroes(elapsed_time):
    """left pad duration with '0' for better alphabetical sorting in html reports"""
    int_decimal = str(round(elapsed_time, 3)).split(".")
    int_decimal[0] = int_decimal[0].zfill(3)
    padded_duration = ".".join(int_decimal)
    return padded_duration


def browser_timestamp_to_datetime(value):
    """Convert a browser timestamp (in milliseconds since epoch) to a datetime object"""
    try:
        timestamp_sec = int(value) / 1000.0
        return datetime.fromtimestamp(timestamp_sec).strftime(
            "%Y-%m-%d %H:%M:%S,%f"
        )[:-3]
    except (ValueError, TypeError):
        return None


def step_text_list_to_html(text):
    """Convert a list of step text lines to an indented HTML heredoc format"""
    text_indent = " " * 8
    heredoc_quote = '"""'
    return "\n".join(
        [text_indent + heredoc_quote]
        + [f"{text_indent}{x}" for x in text]
        + [text_indent + heredoc_quote]
    )


def step_table_to_html(table_data):
    """Convert a step table data structure to an indented HTML table format"""
    text_indent = " " * 8
    return format_gherkin_table(
        table_data["rows"],
        table_data["headings"],
        text_indent,
    )


def generate(results: Path, basepath: Path):
    ## Jinja2 templates setup
    package_loader = jinja2.PackageLoader("cucu.reporter", "templates")
    templates = jinja2.Environment(loader=package_loader)  # nosec
    templates.globals.update(
        escape=escape,
        urlencode=urlencode,
        browser_timestamp_to_datetime=browser_timestamp_to_datetime,
        step_text_list_to_html=step_text_list_to_html,
        step_table_to_html=step_table_to_html,
    )
    feature_template = templates.get_template("feature.html")
    scenario_template = templates.get_template("scenario.html")

    ## prepare report directory
    cucu_dir = Path(sys.modules["cucu"].__file__).parent
    external_dir = cucu_dir / "reporter/external"
    shutil.copytree(external_dir, basepath / "external")
    shutil.copyfile(
        cucu_dir / "reporter/favicon.png",
        basepath / "favicon.png",
    )

    CONFIG.snapshot()

    db_path = results / "run.db"
    try:
        db.init_html_report_db(db_path)

        feature_count = db.feature.select().count()
        scenario_count = db.scenario.select().count()
        step_count = db.step.select().count()
        logger.info(
            f"Starting to process {feature_count} features, {scenario_count} scenarios, and {step_count} steps for report"
        )

        db_features = db.feature.select().order_by(db.feature.start_at)

        features = []
        for db_feature in db_features:
            if db_feature.status == "untested":
                logger.debug(f"Skipping untested feature: {db_feature.name}")
                continue

            feature_dict = shortcuts.model_to_dict(db_feature, backrefs=True)
            features.append(feature_dict)

            feature_results_dir = results
            if db_path := db_feature.worker.cucu_run.db_path:
                logger.debug(
                    f"Combining cucu_runs, using db_path from worker: {db_path}"
                )
                feature_results_dir = Path(db_path).parent

            feature_dict["results_dir"] = feature_results_dir
            feature_dict["folder_name"] = ellipsize_filename(db_feature.name)
            feature_dict["duration"] = (
                feature_dict["start_at"] - feature_dict["start_at"]
            ).total_seconds()

            process_tags(feature_dict)

            feature_path = basepath / feature_dict["folder_name"]

            if feature_dict["status"] not in ["skipped", "untested"]:
                # copy each feature directories contents over to the report directory
                src_feature_filepath = (
                    Path(feature_dict["results_dir"])
                    / feature_dict["folder_name"]
                )

                if src_feature_filepath.exists():
                    shutil.copytree(
                        src_feature_filepath,
                        feature_path,
                        dirs_exist_ok=True,
                    )
                else:
                    logger.warning(
                        f"Feature directory not found, skipping copy: {src_feature_filepath}"
                    )

            db_scenarios = db_feature.scenarios.select().order_by(
                db.scenario.seq
            )

            if len(db_scenarios) == 0:
                logger.debug(f"Feature {db_feature.name} has no scenarios")
                continue

            for scenario_dict in sorted(
                feature_dict["scenarios"], key=lambda x: x["seq"]
            ):
                CONFIG.restore()

                scenario_dict["folder_name"] = ellipsize_filename(
                    scenario_dict["name"]
                )
                scenario_filepath = feature_path / scenario_dict["folder_name"]
                scenario_configpath = (
                    scenario_filepath / "logs/cucu.config.yaml.txt"
                )
                scenario_dict["total_steps"] = len(scenario_dict["steps"])
                if scenario_dict["start_at"]:
                    offset_seconds = (
                        scenario_dict["start_at"] - feature_dict["start_at"]
                    ).total_seconds()
                    scenario_dict["time_offset"] = datetime.fromtimestamp(
                        offset_seconds, timezone.utc
                    )

                if not scenario_configpath.exists():
                    logger.info(f"No config to reload: {scenario_configpath}")
                else:
                    try:
                        CONFIG.load(scenario_configpath)
                    except Exception as e:
                        logger.warning(
                            f"Could not reload config: {scenario_configpath}: {e}"
                        )

                process_tags(scenario_dict)

                sub_headers = []
                for handler in CONFIG[
                    "__CUCU_HTML_REPORT_SCENARIO_SUBHEADER_HANDLER"
                ]:
                    try:
                        sub_header = handler(scenario_dict, feature_dict)
                        if sub_header:
                            sub_headers.append(sub_header)
                    except Exception:
                        logger.warning(
                            f'Exception while trying to run sub_headers hook for scenario: "{scenario_dict["name"]}"\n{traceback.format_exc()}'
                        )
                scenario_dict["sub_headers"] = "<br/>".join(sub_headers)
                scenario_dict["steps"] = sorted(
                    scenario_dict["steps"], key=lambda x: x["seq"]
                )

                for step_dict in scenario_dict["steps"]:
                    # Handle section headings with different levels (# to ####)
                    if step_dict["name"].startswith("#"):
                        # Map the count to the appropriate HTML heading (h2-h5)
                        # We use h2-h5 instead of h1-h4 so h1 can be reserved for scenario/feature titles
                        step_dict["heading_level"] = (
                            f"h{step_dict['name'][:4].count('#') + 1}"
                        )

                    # process timestamps and time offsets
                    if not step_dict["end_at"]:
                        continue

                    if not step_dict["start_at"]:
                        step_dict["timestamp"] = ""
                        step_dict["time_offset"] = ""
                        continue

                    timestamp = step_dict["start_at"]
                    step_dict["timestamp"] = timestamp

                    time_offset = datetime.fromtimestamp(
                        (
                            timestamp - scenario_dict["start_at"]
                        ).total_seconds(),
                        timezone.utc,
                    )
                    step_dict["time_offset"] = time_offset

                logs_path = scenario_filepath / "logs"

                # copy run level console log
                cucu_log_path = behave_filepath_to_cucu_logpath(
                    Path(db_feature.behave_filepath), results
                )
                if cucu_log_path.exists():
                    dest_log_path = logs_path / cucu_log_path.name.lower()
                    dest_log_path.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copyfile(cucu_log_path, dest_log_path)

                log_files = []
                for log_file in logs_path.glob("*.*"):
                    log_filepath = log_file.relative_to(scenario_filepath)

                    if (
                        scenario_dict["start_at"]
                        and ".console." in log_filepath.name
                    ):
                        log_filepath = Path(f"logs/{log_filepath.name}.html")

                    log_files.append(
                        {
                            "filepath": log_filepath,
                            "name": log_file.name,
                        }
                    )

                # generate html version of console log
                for log_file in [
                    x for x in log_files if ".console." in x["name"]
                ]:
                    input_file = scenario_filepath / "logs" / log_file["name"]
                    output_file = scenario_filepath / log_file["filepath"]
                    output_file.write_text(
                        parse_log_to_html(
                            input_file.read_text(encoding="utf-8")
                        ),
                        encoding="utf-8",
                    )

                scenario_dict["logs"] = log_files

                # render scenario html
                scenario_basepath = feature_path / scenario_dict["folder_name"]
                scenario_basepath.mkdir(parents=True, exist_ok=True)
                rendered_scenario_html = scenario_template.render(
                    basepath=results,
                    feature=feature_dict,
                    path_exists=lambda path: Path(path).exists(),
                    scenario=scenario_dict,
                    steps=scenario_dict["steps"],
                    title=scenario_dict["name"],
                    dir_depth="../../",
                )
                scenario_output_filepath = scenario_basepath / "index.html"
                scenario_output_filepath.write_text(rendered_scenario_html)

            # render feature html
            rendered_feature_html = feature_template.render(
                feature=feature_dict,
                scenarios=feature_dict["scenarios"],
                dir_depth="",
                title=feature_dict["name"],
            )
            feature_output_filepath = basepath / f"{feature_dict['name']}.html"
            feature_output_filepath.write_text(rendered_feature_html)

            feature_dict["total_steps"] = sum(
                [x["total_steps"] for x in feature_dict["scenarios"]]
            )
            feature_dict["duration"] = left_pad_zeroes(
                sum(
                    [
                        float(x["duration"])
                        for x in feature_dict["scenarios"]
                        if x["duration"]
                    ]
                )
            )

        # query the database for stats
        feature_stats_db = db.db.execute_sql("SELECT * FROM flat_feature")
        keys = tuple([x[0] for x in feature_stats_db.description])
        feature_stats = [
            dict(zip(keys, x)) for x in feature_stats_db.fetchall()
        ]

        grand_totals_db = db.db.execute_sql("SELECT * FROM flat_all")
        keys = tuple([x[0] for x in grand_totals_db.description])
        grand_totals = dict(zip(keys, grand_totals_db.fetchone()))

        ## Generate index.html and flat.html

        index_template = templates.get_template("index.html")
        rendered_index_html = index_template.render(
            feature_stats=feature_stats,
            grand_totals=grand_totals,
            title="Cucu HTML Test Report",
            basepath=basepath,
            dir_depth="",
        )
        html_index_path = basepath / "index.html"
        html_index_path.write_text(rendered_index_html)

        flat_template = templates.get_template("flat.html")
        rendered_flat_html = flat_template.render(
            features=features,
            grand_totals=grand_totals,
            title="Flat HTML Test Report",
            basepath=basepath,
            dir_depth="",
        )
        html_flat_path = basepath / "flat.html"
        html_flat_path.write_text(rendered_flat_html)

    finally:
        db.close_html_report_db()

    return html_flat_path



import contextlib
import io
import json
import re

from cucu import behave_tweaks


def load_cucu_steps(filepath=None):
    """
    loads the cucu steps definition using behave and returns an array of
    hashmaps that have the following structure:

        {
            [step_name]: {
                "location": {
                    "filepath": "...",
                    "line": "...",
                }
            }
        }

    undefined steps are marked with a value of None instead of a location. ie:

        {
            [undefined_step_name]: None
        }

    Returns:
        an array of hashmaps
    """
    steps_cache = {}
    args = ["--dry-run", "--no-summary", "--format", "steps.doc"]

    if filepath is not None:
        args.append(filepath)

    error = None

    stdout = io.StringIO()
    stderr = io.StringIO()

    with contextlib.redirect_stderr(stderr):
        with contextlib.redirect_stdout(stdout):
            error = behave_tweaks.behave_main(args)

    stdout = stdout.getvalue()
    stderr = stderr.getvalue()

    if stdout.startswith("USING RUNNER:"):
        # remove the first line of the output so that linter won't error out trying to parse it as a step
        stdout = "\n".join(stdout.split("\n")[1:])

    if stdout.startswith("ParserError"):
        print(stdout)
        raise RuntimeError(
            "unable to parse feature files, see above for details"
        )

    for cucu_step in stdout.split("@step"):
        # Each line of a step definition looks like so:
        #
        #   @step('I should see "{this}" matches "{that}"')
        #     Function: inner_step()
        #     Location: src/cucu/behave_tweaks.py:64
        #      possibly a doc string here on the function that can be used
        #      to add a little documentation to each step definition
        #
        #   @step('I should see "{this}" matches the following')
        #     Function: inner_step()
        #     Location: src/cucu/behave_tweaks.py:64
        #
        if cucu_step.strip() == "":
            continue

        if not cucu_step.startswith("("):
            #
            # any block of lines between the `@step` that doesn't start with the
            # character ( is an error being reported behave when loading steps
            # and we'll ignore it when processing the step definitions and then
            # report the actual underlying trace reported in STDERR below
            #
            print("unable to parse some step lines")
            continue

        lines = cucu_step.split("\n")

        #
        # parts[1] is the function name while parts[3:] is the docstring
        # of the step which we can use for documenting usage of the step
        # in the language server
        #
        step_name = lines[0]
        location = lines[2]

        step_name = re.match(r"\('(.*)'\)", step_name).groups()[0]
        _, filepath, line_number = location.split(":")

        steps_cache[step_name] = {
            "location": {
                "filepath": filepath.strip(),
                "line": line_number.strip(),
            }
        }

    # collect any undefined steps
    for line in stderr.split("\n"):
        # @when(u'I close the current browser caca')
        match = re.match(r"@[a-z]+\(u'([^']+)'\)", line)
        if match:
            step_name = match.group(1)
            steps_cache[step_name] = None

    if error == 0:
        return (steps_cache, None)

    else:
        return (steps_cache, stderr)


def print_json_steps(filepath=None):
    """
    pretty print the steps in a JSON fart
    """
    steps, steps_error = load_cucu_steps(filepath=filepath)
    print(json.dumps(steps, indent=2, sort_keys=True))


def print_human_readable_steps(filepath=None):
    steps, steps_error = load_cucu_steps(filepath=filepath)

    for step_name in steps:
        if steps[step_name] is not None:
            if str(filepath) in steps[step_name]["location"]["filepath"]:
                print(f"custom: {step_name}")
            else:
                print(f"cucu:   {step_name}")

    if steps_error is not None:
        print(steps_error)
        raise RuntimeError("Failure loading some steps, see above for details")



import html
import re
from html import escape

from behave.formatter.ansi_escapes import colors, escapes

ESC_SEQ = r"\x1b["
TRANSLATION = (
    {colors["grey"]: '<span style="color: grey;">\n'}
    | {  # grey indicates a new step, and hence add a line break
        v: f'<span style="color: {k};">'
        for k, v in colors.items()
        if k != "grey"
    }
    | {
        "\n" + escapes["reset"] + "\n": "</span>\n",
        escapes["reset"]: "</span>",
        escapes["up"] + "\n": "",  # remove a line break with cursor up
        ESC_SEQ + "46": "",  # ignore DELETE (num keypad)
        ESC_SEQ + "48": "",  # ignore INSERT (num keypad)
        ESC_SEQ + "49": "",  # ignore END (num keypad)
        ESC_SEQ + "50": "",  # ignore DOWN ARROW (num keypad)
        ESC_SEQ + "51": "",  # ignore PAGE DOWN (num keypad)
        ESC_SEQ + "52": "",  # ignore LEFT ARROW (num keypad)
        ESC_SEQ + "54": "",  # ignore RIGHT ARROW (num keypad)
        ESC_SEQ + "55": "",  # ignore HOME (num keypad)
        ESC_SEQ + "56": "",  # ignore UP ARROW (num keypad)
        ESC_SEQ + "57": "",  # ignore PAGE UP (num keypad)
        ESC_SEQ + "1A": "",  # ignore move cursor up one line
    }
)
RE_TO_HTML = re.compile("|".join(map(re.escape, TRANSLATION)))

RE_TO_REMOVE = re.compile(
    r"\x1b\[(0;)?[0-9A-F]{1,2}[mA]:?"
)  # detect hex values, not just decimal digits


def remove_ansi(input: str) -> str:
    return RE_TO_REMOVE.sub("", input)


def parse_log_to_html(input: str) -> str:
    """
    Parse an ansi color log to html
    """
    html_start = "<!DOCTYPE html>\n<html>\n"
    html_end = "</html>\n"
    head = '<head>\n<meta http-equiv="Content-Type" content="text/html; charset=utf-8">\n</head>\n'
    body_start = '<body style="color: white; background-color: #100c08;">'  # use dark bg since colors are from behave
    body_end = "</body>\n"
    result = f"{html_start}{head}{body_start}<pre>\n{RE_TO_HTML.sub(lambda match: TRANSLATION[match.group(0)], html.escape(input, quote=False))}\n</pre>{body_end}{html_end}"
    if ESC_SEQ in result:
        lines = "\n".join([x for x in result.split("\n") if ESC_SEQ in x])

        print(
            f"Detected unmapped ansi escape code!:\n{lines}"
        )  # use print instead of logger to avoid circular import

    return result


ANSI_CLASSES = {
    "0": [],  # reset
    "1": ["ansi-bold"],
    "30": ["ansi-black"],
    "31": ["ansi-red"],
    "32": ["ansi-green"],
    "33": ["ansi-yellow"],
    "34": ["ansi-blue"],
    "35": ["ansi-magenta"],
    "36": ["ansi-cyan"],
    "37": ["ansi-white"],
    "90": ["ansi-gray"],
}

ANSI_SGR_RE = re.compile(r"\x1b\[([0-9;]+)m")


def ansi_to_html(line: str) -> str:
    """
    Convert ANSI SGR codes to <span class="..."> HTML.
    """
    result = []
    open_spans = []

    pos = 0
    for match in ANSI_SGR_RE.finditer(line):
        chunk = line[pos : match.start()]
        if chunk:
            result.append(escape(chunk))

        codes = match.group(1).split(";")

        # reset
        if "0" in codes:
            while open_spans:
                result.append("</span>")
                open_spans.pop()
        else:
            classes = []
            for c in codes:
                classes.extend(ANSI_CLASSES.get(c, []))

            if classes:
                class_attr = " ".join(classes)
                result.append(f'<span class="{class_attr}">')
                open_spans.append("</span>")

        pos = match.end()

    # remainder
    remainder = line[pos:]
    if remainder:
        result.append(escape(remainder))

    # close any open spans
    while open_spans:
        result.append(open_spans.pop())

    return "".join(result)



import re

from cucu import config, step
from cucu.steps import step_utils


@step('I set the variable "{variable}" to "{value}"')
def set_variable_to(_, variable, value):
    config.CONFIG[variable] = value


@step('I set the variable "{variable}" to the following')
def set_variable_to_the_following(ctx, variable):
    config.CONFIG[variable] = ctx.text


@step('I should see "{this}" is empty')
def should_see_is_empty(_, this):
    if this or len(this) != 0:
        raise RuntimeError(f"{this} is not empty")


@step('I should see "{this}" is equal to "{that}"')
def should_see_is_equal(_, this, that):
    if this != that:
        raise RuntimeError(f"{this} is not equal to {that}")


@step('I should see "{this}" is not equal to "{that}"')
def should_see_is_not_equal(_, this, that):
    if this == that:
        raise RuntimeError(f"{this} is equal to {that}")


@step('I should see "{this}" contains "{that}"')
def should_see_it_contains(_, this, that):
    if that not in this:
        raise RuntimeError(f"{this} does not contain {that}")


@step('I should see "{this}" contains the following')
def should_see_it_contains_the_following(ctx, this):
    if ctx.text not in this:
        raise RuntimeError(f"{this} does not contain {ctx.text}")


@step('I should see "{this}" does not contain "{that}"')
def should_see_it_doest_not_contain(_, this, that):
    if that in this:
        raise RuntimeError(f"{this} contains {that}")


@step('I should see "{this}" does not contain the following')
def should_see_it_does_not_contain(ctx, this):
    if ctx.text in this:
        raise RuntimeError(f"{this} contain {ctx.text}")


@step('I should see "{this}" is equal to the following')
def should_see_is_equal_to_the_following(ctx, this):
    that = ctx.text

    if this != that:
        raise RuntimeError(f"{this} is not equal to {that}")


@step('I should see "{this}" matches "{that}"')
def should_see_matches(_, this, that):
    if re.match(that, this) is None:
        raise RuntimeError(f"{this} does not match {that}")


@step('I should see "{this}" matches the following')
def should_see_matches_the_following(ctx, this):
    that = ctx.text

    if re.match(that, this) is None:
        raise RuntimeError(f"{this}\ndoes not match:\n{that}")


@step('I should see "{this}" does not match the following')
def should_does_not_see_matches_the_following(ctx, this):
    that = ctx.text

    if re.match(that, this) is not None:
        raise RuntimeError(f"{this}\nmatches:\n{that}")


@step(
    'I search for the regex "{regex}" in "{value}" and save the group "{name}" to the variable "{variable}"'
)
def search_and_save(ctx, regex, value, name, variable):
    step_utils.search_and_save(regex, value, name, variable)


@step(
    'I match the regex "{regex}" in "{value}" and save the group "{name}" to the variable "{variable}"'
)
def match_and_save(ctx, regex, value, name, variable):
    step_utils.match_and_save(regex, value, name, variable)



from cucu import fuzzy, helpers
from cucu.utils import (
    find_n_click_input_parent_label,
    is_element_size_zero,
    take_saw_element_screenshot,
)

from . import base_steps


def find_button(ctx, name, index=0):
    """
    find a button on screen by fuzzy matching on the name and index provided.

        * <button>
        * <input type="button">
        * <input type="submit">
        * <a>
        * <* role="button">
        * <* role="link">
        * <* role="menuitem">
        * <* role="treetem">
        * <* role="option">
        * <* role="radio">

    note: the reason we're allowing other items such as menuitem, option, etc
          is that on screen they can present themselves like "buttons". When
          searching for more things to include use the following image
          reference:

          https://www.w3.org/TR/2009/WD-wai-aria-20091215/rdf_model.png

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired button on screen
      index(str):  the index of the button if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()
    element = fuzzy.find(
        ctx.browser,
        name,
        [
            "button",
            'input[type="button"]',
            'input[type="submit"]',
            "a",
            '*[role="button"]',
            '*[role="link"]',
            '*[role="menuitem"]',
            '*[role="treeitem"]',
            '*[role="option"]',
            '*[role="radio"]',
        ],
        index=index,
    )

    take_saw_element_screenshot(ctx, "button", name, index, element)

    return element


def click_button(ctx, button):
    """
    internal method used to simply click a button element
    """
    ctx.check_browser_initialized()

    if base_steps.is_disabled(button):
        raise RuntimeError("unable to click the button, as it is disabled")

    if is_element_size_zero(button):
        find_n_click_input_parent_label(ctx, button)
        return

    ctx.browser.click(button)


helpers.define_should_see_thing_with_name_steps(
    "button", find_button, with_nth=True
)
helpers.define_action_on_thing_with_name_steps(
    "button", "click", find_button, click_button, with_nth=True
)
helpers.define_thing_with_name_in_state_steps(
    "button", "disabled", find_button, base_steps.is_disabled, with_nth=True
)
helpers.define_thing_with_name_in_state_steps(
    "button",
    "not disabled",
    find_button,
    base_steps.is_not_disabled,
    with_nth=True,
)
helpers.define_run_steps_if_I_can_see_element_with_name_steps(
    "button", find_button
)
